/*
 * bloom_math: A program to do math with bloom filters.
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include "config.h"
#include "bloom.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <errno.h>
#include <err.h>
#include <getopt.h>
#include <ctype.h>
#include <getopt.h>
#include <ctype.h>
#include <math.h>

#ifdef HAVE_COPYFILE_H
#include <copyfile.h>
#endif

const char *progname = "bloomcat";

void usage()
{
    printf("Usage: %s [options] bloomfile1  [bf2 bf3...]\n",progname);
    printf("Do math with one or more bloom files \n");
    printf("     -o outfile   Arithmatic 'or' with output to outfile.\n");
    printf("     -c           Reports correlation between multiple bloom filters (# out of #)\n");
    exit(0);
}

#ifndef HAVE_COPYFILE_H
/* Simson's cut-rate copyfile implementation */
#define COPYFILE_ALL 0
int fcopyfile(int ifd,int ofd,int f1,int f2)
{
    char buf[65536];
    int r;
    while((r = read(ifd,buf,sizeof(buf)))>0){
	if(write(ofd,buf,r)!=r) return(-1);
    }
    return 0;
}
#endif


/* Code to quickly and efficiently count the number of 1's in an unsigned char.
 * See http://stackoverflow.com/questions/697978/c-code-to-count-the-number-of-1-bits-in-an-unsigned-char
 */
const unsigned char oneBits[] = {0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4};
inline unsigned char CountOnes(unsigned char x)
{
    unsigned char results;
    results = oneBits[x&0x0f];
    results += oneBits[x>>4];
    return results;
}


#define min(x,y) ((x)<(y)?(x):(y))

int main(int argc,char **argv)
{
    const char *outfile = 0;
    progname = argv[0];
    int opt_c = 0;
    int ch;
    
    while ((ch = getopt(argc,argv,"co:h?")) != -1){
	switch(ch){
	case 'c':
	    opt_c = 1;break;
	case 'o':
	    outfile = optarg; break;
	default:
	case '?':
	case 'h':
	    usage(); exit(0);
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<1) usage();

    /* Get the parameters for the first bloom file and compare it to the others, make
     * sure they are all comsurate.
     */

    NSRLBloom b1;
    if(b1.open(argv[0],O_RDONLY)) err(1,"%s",argv[0]);
    
    /* Check all of the other blooms... */
    for(int i=1;i<argc;i++){
	NSRLBloom bn;
	if(bn.open(argv[i],O_RDONLY)) err(1,"%s",argv[i]);
	if(b1.hash_bytes != bn.hash_bytes) err(1,"%s: hash_bytes %d!=%d",argv[i],b1.hash_bytes,bn.hash_bytes);
	if(b1.M != bn.M) err(1,"%s: bloom_bits %d!=%d",argv[i],b1.M,bn.M);
	if(b1.k != bn.k) err(1,"%s: k %d!=%d",argv[i],b1.k,bn.k);
	if(b1.vector_bytes != bn.vector_bytes){
	    errx(1,"%s: vector_bytes %zd!=%zd",argv[i],b1.vector_bytes,bn.vector_bytes);
	}
    }

    if(opt_c){
	/* Compute the correlation */
	if(argc!=2) err(1,"-c requires two bloom filters\n");

	NSRLBloom b2;
	if(b2.open(argv[1],O_RDONLY)) err(1,"%s",argv[1]);

	if(b1.k != b2.k) err(1,"b1 and b2 have different k! (%d != %d)\n",b1.k,b2.k);
	int k = b1.k;

	uint64_t e_1 = 0;		// total bits set in filter 1
	uint64_t e_2 = 0;		// total bits set in filter 2
	uint64_t e_12 = 0;		// total bits set in filter 1 & 2
	uint64_t m = b1.vector_bytes*8; // number of bits total

	uint64_t s_1 = b1.unique_added_items;
	uint64_t s_2 = b2.unique_added_items;
	uint64_t s_12 = 0;		// number of common elements; what we are trying to estimate

	for(uint32_t j=0;j<b2.vector_bytes;j++){
	    e_1  += CountOnes(b1.vector[j]);
	    e_2  += CountOnes(b2.vector[j]);
	    e_12 += CountOnes(b1.vector[j] & b2.vector[j]);
	}
	double p = 1.0 - 1.0/m;
	double Emin = m * (1 - pow(p,k * s_1) - pow(p,k * s_2) + pow(p,k*(s_1+s_2)));
	double Emax = min(e_1,e_2);
	double E_12 = m * (1 - pow(p,k * s_1) - pow(p,k * s_2) + pow(p,k*(s_1+s_2-s_12)));
	double alpha = 0.3;
	double C = alpha * (Emax-Emin) + Emin;
	double SF = 100 * (e_12-C) / (Emax-C);

	if (e_12<C) SF = 0;

	printf("s_1: %8"PRId64,s_1);
	printf("s_2: %8"PRId64,s_2);
	printf(" Emin: %6f  Emax: %6f  E_12: %6f  SF: %6f\n", Emin,Emax,E_12,SF);
	return(0);
    }

    if(outfile){
	/* Looks good. Now add them together for real */
	int ofd = open(outfile,O_RDWR|O_CREAT|O_EXCL,0666);
	if(ofd<0) err(1,"%s",outfile);
	
	/* Copy the first file! */
	lseek(b1.fd,0,SEEK_SET);
	lseek(ofd,0,SEEK_SET);
	if(fcopyfile(b1.fd,ofd,0,COPYFILE_ALL)) err(1,"fcopyfile");
	close(ofd);
	
	/* Open the copy */
	NSRLBloom bc;
	if(bc.open(outfile,O_RDWR)) err(1,"%s",outfile);
	
	/* Do the "or" */
	for(int i=1;i<argc;i++){
	    NSRLBloom bn;
	    if(bn.open(argv[i],O_RDONLY)) err(1,"%s",argv[i]);
	    for(uint32_t j = 0; j<bn.vector_bytes;j++){
		bc.vector[j] |= bn.vector[j];
	    }
	}
    }
    exit(0);
}




