/**
 * discriminator.h:
 * defines the abstract discriminator class and some specific discriminators.
 */

#ifndef DISCRIMINATOR_H
#define DISCRIMINATOR_H

#include <sys/types.h>
#include <vector>
#include <iostream>

#include "myregex.h"
#include "histogram.h"
#include "sdbuf.h"

/* specificity ranges
 *  9 - JPEG header, JPEG body, NULL
 *  5 - HUFFMAN-coded
 *  2 - ASCII
 *  1 - high entropy, low entropy
 */

using namespace std;

class vote {
 public:
 vote(int64_t aBlock,string aName,string aDetail,double aTPR,double aFPR,int aSpec):
    block_number(aBlock),name(aName),details(aDetail),TPR(aTPR),FPR(aFPR),spec(aSpec){}
    int64_t block_number;			/* which block are we voting on? */
    string name;
    string details;			/* how did we find it */
    double TPR;				/* true positive rate */
    double FPR;				/* false positive rate */
    int spec;				/* specificity; 1-100*/

    friend std::ostream & operator<<(std::ostream &os,const vote &v){
	os << v.block_number << " " << v.name << " " << v.details;
	return os;
    }
};

class votelist_t: public vector<vote>{
 public:;
    friend std::ostream & operator<<(std::ostream &os,const votelist_t &vl){
	for(votelist_t::const_iterator it = vl.begin(); it!=vl.end(); it++){
	    cout << " " << (*it) << "\n";
	}
	return os;
    }
    /**
     * Find the best vote, which is the vote with the highest specificity, then
     * the lowest false positive rate, then the highest true positive rate.
     */
    const vote *best_vote(){
	const vote *best = 0;
	for(votelist_t::const_iterator it = this->begin(); it!=this->end(); it++){
	    const vote &v = (*it);
	    if((best==0) ||
	       (best->spec < v.spec) ||
	       ((best->spec == v.spec) && (best->FPR > v.FPR)) ||
	       ((best->spec == v.spec) && (best->FPR == v.FPR) && (best->TPR < v.TPR))){
		best = &(*it);
	    }
	}
	return best;
    }
};

class discriminator {
 public:
    virtual string name()=0;
    virtual votelist_t identify(const sdbuf &buf)=0;
    virtual bool notable_votes(){return false;} /* should individual votes be kept? */
    virtual ~discriminator(){};
};

class discriminator_list : public vector<discriminator *> {
 public:
    static void init();			/* initializes discriminator list */
};
extern discriminator_list discriminators;


class null_discriminator:public discriminator {
public:
    virtual ~null_discriminator(){}
    null_discriminator(){};
    virtual string name(){return string("NULL");}
    virtual votelist_t identify(const sdbuf &buf);
};

class jpeg_discriminator:public discriminator {
public:
    virtual ~jpeg_discriminator(){}
    static const int JPEG_HE=200;
    static const int JPEG_LN=2;
    jpeg_discriminator(){};
    virtual string name(){return string("JPEG");}
    virtual votelist_t identify(const sdbuf &buf);
};

class entropy_discriminator:public discriminator {
 public:
    virtual ~entropy_discriminator(){}
    static const int ENTROPY_HIGH=200;
    static const int ENTROPY_LOW=10;
    static const int MAX_NON_ASCII=10;
    entropy_discriminator(){};
    virtual string name(){return string("ENTROPY");}
    virtual votelist_t identify(const sdbuf &buf);
};

class huffman_discriminator:public discriminator {
 public:
    huffman_discriminator(){};
    virtual ~huffman_discriminator(){};
    virtual string name(){return string("HUFFMAN");}
    virtual votelist_t identify(const sdbuf &buf);
};

class pdf_discriminator:public discriminator {
 public:
    myregex *reg1;
    pdf_discriminator(){
	reg1 = 0;
    };
    virtual string name(){return string("PDF");}
    virtual votelist_t identify(const sdbuf &buf);
};

class bloom_discriminator:public discriminator {
 private:
    class NSRLBloom *bf;
    string fname;
 public:
    static bloom_discriminator *new_bloom(string fn);
    bloom_discriminator(){};
    virtual string name(){return string("BLOOM");}
    virtual votelist_t identify(const sdbuf &buf);
    virtual bool notable_votes(){return true;}
};

#endif
