/*
 * Campare the speed of various hash functions in C++
 */

#include "config.h"

#include "bloom.h"

#include <stdio.h>
#include <sys/types.h>
#include <openssl/hmac.h>
#include <openssl/rand.h>
#include <string.h>
#include <cstring>
#include "aftimer.h"
#include "crc64.h"
#include "adler32.h"

#ifdef HAVE_LOOKUP3_H
#include "lookup3.h"
#endif

void hashtest(size_t insize)
{
    int COUNT=1000000*512/insize+25000;
    const char *tests[] = {"Adler32","Lookup3","CRC64","MD5","SHA1","SHA256",0};
    u_char *inbuf = (u_char *)alloca(insize);
    memset((void *)inbuf,'E',insize);

    printf("=== C++ -- %8d x %6zu-byte blocks ===\n\n",COUNT,insize);
    for(int j=0;tests[j];j++){
#ifndef HAVE_LOOKUP3
	if(strcmp(tests[j],"Lookup3")==0) continue;
#endif	    
	printf("%7s ...",tests[j]);
	fflush(stdout);
	aftimer t;
	unsigned int r = 0;
	u_char out[EVP_MAX_MD_SIZE];
	u_int outlen = sizeof(out);

	t.start();
	for(int i=0;i<COUNT;i++){
	    inbuf[0] = i%256;		// change the input
	    switch(j){
	    case 0: /* Adler 32 */      r = av_adler32_update(1,inbuf,insize);			break;
#ifdef HAVE_LOOKUP3
	    case 1: /* Lookup3  */      r = hashlittle(inbuf,insize,0);				break;
#endif
	    case 2: /* CRC64 */         crc64(inbuf,insize,out);				break;
	    case 3: /* MD5 */		EVP_Digest(inbuf,insize,out,&outlen,EVP_md5(),0);	break;
	    case 4: /* SHA1 */		EVP_Digest(inbuf,insize,out,&outlen,EVP_sha1(),0);	break;
	    case 5: /* SHA256 */        EVP_Digest(inbuf,insize,out,&outlen,EVP_sha256(),0);	break;
	    }
	}
	t.stop();
	if(r==0) r = out[0] | (out[1]<<8) | (out[2]<<16) | (out[3]<<24);
	printf("time: %8.2f sec; %8.1f Khashes/sec; %8.1f Mb/sec r=%08x\n",
	       t.elapsed_seconds(),
	       ((double)COUNT)/t.elapsed_seconds()/1000.0,
	       ((double)COUNT)*insize/t.elapsed_seconds()/1000000.0,
	       r
	       );
    }

    printf("---------------\n\n");
}


int main(int argc,char **argv)
{
    hashtest(512);
    hashtest(4096);
    hashtest(65536);

    return(0);
}

